from dataclasses import dataclass

import pytest

from richset import RichSet


@dataclass(frozen=True)
class Something:
    id: int
    name: str


def test_richset_pushed() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )
    assert rs.pushed(Something(4, "four")).to_list() == [
        Something(1, "one"),
        Something(2, "two"),
        Something(3, "three"),
        Something(4, "four"),
    ]


def test_richset_pushed_all() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )
    assert rs.pushed_all([]).to_list() == [
        Something(1, "one"),
        Something(2, "two"),
        Something(3, "three"),
    ]

    assert rs.pushed_all(
        [
            Something(4, "four"),
            Something(5, "five"),
        ]
    ).to_list() == [
        Something(1, "one"),
        Something(2, "two"),
        Something(3, "three"),
        Something(4, "four"),
        Something(5, "five"),
    ]


def test_richset_unshifted() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )
    assert rs.unshifted(Something(4, "four")).to_list() == [
        Something(4, "four"),
        Something(1, "one"),
        Something(2, "two"),
        Something(3, "three"),
    ]


def test_richset_unshifted_all() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )
    assert rs.unshifted_all([]).to_list() == [
        Something(1, "one"),
        Something(2, "two"),
        Something(3, "three"),
    ]

    assert rs.unshifted_all(
        [
            Something(4, "four"),
            Something(5, "five"),
        ]
    ).to_list() == [
        Something(4, "four"),
        Something(5, "five"),
        Something(1, "one"),
        Something(2, "two"),
        Something(3, "three"),
    ]


def test_richset_popped() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )

    item, rs2 = rs.popped()
    assert item == Something(3, "three")
    assert rs2.to_list() == [
        Something(1, "one"),
        Something(2, "two"),
    ]

    rs = RichSet.from_list([])
    with pytest.raises(IndexError) as err:
        rs.popped()
    assert str(err.value) == "pop from empty RichSet"


def test_richset_popped_n() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )

    popped_items, rs2 = rs.popped_n(2)
    assert popped_items.to_list() == [
        Something(3, "three"),
        Something(2, "two"),
    ]
    assert rs2.to_list() == [Something(1, "one")]

    rs.popped_n(3)
    with pytest.raises(IndexError) as err:
        rs.popped_n(4)
    assert str(err.value) == "pop more than size"

    rs = RichSet.from_list([])
    with pytest.raises(IndexError) as err:
        rs.popped_n(2)
    assert str(err.value) == "pop more than size"


def test_richset_shifted() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )

    item, rs2 = rs.shifted()
    assert item == Something(1, "one")
    assert rs2.to_list() == [
        Something(2, "two"),
        Something(3, "three"),
    ]

    rs = RichSet.from_list([])
    with pytest.raises(IndexError) as err:
        rs.shifted()
    assert str(err.value) == "shift from empty RichSet"


def test_richset_shifted_n() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )

    shifted_items, rs2 = rs.shifted_n(2)
    assert shifted_items.to_list() == [
        Something(1, "one"),
        Something(2, "two"),
    ]
    assert rs2.to_list() == [Something(3, "three")]

    rs.shifted_n(3)
    with pytest.raises(IndexError) as err:
        rs.shifted_n(4)
    assert str(err.value) == "shift more than size"

    rs = RichSet.from_list([])
    with pytest.raises(IndexError) as err:
        rs.shifted_n(2)
    assert str(err.value) == "shift more than size"


def test_richset_slice() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )
    assert rs.slice(0, 0).to_list() == []
    assert rs.slice(1, 2).to_list() == [Something(2, "two")]
    assert rs.slice(0, 3).to_list() == rs.to_list()
    assert rs.slice(0, 4).to_list() == rs.to_list()
    assert rs.slice(1, 3).to_list() == [
        Something(2, "two"),
        Something(3, "three"),
    ]
    assert rs.slice(-1, 3).to_list() == [Something(3, "three")]


def test_richset_divide_at() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )
    rs2, rs3 = rs.divide_at(0)
    assert rs2.to_list() == []
    assert rs3.to_list() == rs.to_list()

    rs2, rs3 = rs.divide_at(1)
    assert rs2.to_list() == [Something(1, "one")]
    assert rs3.to_list() == [Something(2, "two"), Something(3, "three")]

    rs2, rs3 = rs.divide_at(2)
    assert rs2.to_list() == [Something(1, "one"), Something(2, "two")]
    assert rs3.to_list() == [Something(3, "three")]

    rs2, rs3 = rs.divide_at(3)
    assert rs2.to_list() == [
        Something(1, "one"),
        Something(2, "two"),
        Something(3, "three"),
    ]
    assert rs3.to_list() == []



from dataclasses import dataclass

from richset import RichSet


@dataclass(frozen=True)
class Something:
    id: int
    name: str


def test_richset_from_list() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs == RichSet.from_list([Something(1, "one"), Something(2, "two")])



from dataclasses import dataclass

from richset import RichSet


@dataclass(frozen=True)
class Something:
    id: int
    name: str


def test_richset_is_empty() -> None:
    assert RichSet.from_list([]).is_empty()
    assert RichSet[str].from_empty().is_empty()
    assert not RichSet.from_list([Something(1, "one")]).is_empty()


def test_richset_is_non_empty() -> None:
    assert RichSet.from_list([]).is_non_empty() is False
    assert RichSet[str].from_empty().is_non_empty() is False
    assert RichSet.from_list([Something(1, "one")]).is_non_empty() is True


def test_richset_size() -> None:
    assert RichSet[str].from_empty().size() == 0
    assert RichSet.from_list([Something(1, "one")]).size() == 1
    assert (
        RichSet.from_list(
            [
                Something(1, "one"),
                Something(2, "two"),
            ]
        ).size()
        == 2
    )


def test_richset_count() -> None:
    assert RichSet.from_list([]).count(lambda r: True) == 0
    assert RichSet.from_list([]).count(lambda r: False) == 0
    rs = RichSet.from_list([Something(1, "one"), Something(2, "two")])
    assert rs.count(lambda r: r.id == 1) == 1
    assert rs.count(lambda r: r.id > 0) == 2
    assert (
        RichSet.from_list(
            [
                Something(1, "one"),
                Something(2, "two"),
                Something(3, "three"),
            ]
        ).count(lambda r: r.id % 2 == 0)
        == 1
    )



from __future__ import annotations

import functools
from dataclasses import dataclass
from typing import (
    Callable,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    Literal,
    TypeVar,
    overload,
)

from .comparable import Comparable

T = TypeVar("T")
S = TypeVar("S")
Key = TypeVar("Key", bound=Hashable)


@dataclass(frozen=True)
class RichSet(Generic[T]):
    records: tuple[T, ...]

    # factory classmethods

    @classmethod
    def from_list(cls, lst: list[T]) -> RichSet[T]:
        """Returns a new RichSet from a list."""
        return cls.from_iterable(lst)

    @classmethod
    def from_tuple(cls, tpl: tuple[T, ...]) -> RichSet[T]:
        """Returns a new RichSet from a tuple."""
        return cls.from_iterable(tpl)

    @classmethod
    def from_set(cls, s: set[T]) -> RichSet[T]:
        """Returns a new RichSet from a set."""
        return cls.from_iterable(s)

    @classmethod
    def from_frozenset(cls, s: frozenset[T]) -> RichSet[T]:
        """Returns a new RichSet from a frozenset."""
        return cls.from_iterable(s)

    @classmethod
    def from_iterable(cls, itr: Iterable[T]) -> RichSet[T]:
        """Returns a new RichSet from an iterable."""
        return cls(tuple(itr))

    @classmethod
    def from_empty(cls) -> RichSet[T]:
        """Returns an empty RichSet."""
        return cls(records=())

    # magic methods

    def __iter__(self) -> Iterator[T]:
        return iter(self.records)

    def __len__(self) -> int:
        return self.size()

    # conversions

    def to_list(self) -> list[T]:
        """Returns a list of records."""
        return list(self.records)

    def to_tuple(self) -> tuple[T, ...]:
        """Returns a tuple of records."""
        return tuple(self.records)

    def to_set(
        self,
    ) -> set[T]:  # intersection of T & Hashable
        """Returns a set of records."""
        for r in self.records:
            if not isinstance(r, Hashable):
                raise TypeError(f"non-hashable record: {r}")
        return set(self.records)

    def to_frozenset(
        self,
    ) -> frozenset[T]:  # intersection of T & Hashable
        """Returns a frozenset of records."""
        return frozenset(self.to_set())

    def to_dict(
        self,
        key: Callable[[T], Key],
        *,
        duplicated: Literal["error", "first", "last"] = "error",
    ) -> dict[Key, T]:
        """Returns a dictionary mapping keys to values.

        if duplicated is 'error' (default), raises an error if \
there are multiple records with the same key.
        if duplicated is 'first', returns the first record with the same key.
        if duplicated is 'last', returns the last record with the same key."""
        base = self.to_dict_of_list(key=key)

        if duplicated == "error":
            if len(base) != self.size():
                raise ValueError("duplicate keys")
            return {k: v[0] for k, v in base.items()}
        elif duplicated == "first":
            return {k: v[0] for k, v in base.items()}
        elif duplicated == "last":
            return {k: v[-1] for k, v in base.items()}
        else:
            raise ValueError("invalid duplicated value")

    def to_dict_of_list(
        self,
        key: Callable[[T], Key],
    ) -> dict[Key, list[T]]:
        """Returns a dictionary mapping keys to lists of values."""
        d: dict[Key, list[T]] = {}
        for r in self.records:
            k = key(r)
            if k not in d:
                d[k] = []
            d[k].append(r)
        return d

    # list accessors

    @overload
    def get_first(self) -> T | None:
        ...

    @overload
    def get_first(self, default: S) -> T | S:
        ...

    def get_first(self, default: S | None = None) -> T | S | None:
        """Returns the first record in the RichSet
        or default value (None) if the RichSet is empty."""
        if self.records:
            return self.records[0]
        return default

    def first(self) -> T:
        """Returns the first record in the RichSet."""
        if self.records:
            return self.records[0]
        raise IndexError("RichSet is empty")

    @overload
    def get_last(self) -> T | None:
        ...

    @overload
    def get_last(self, default: S) -> T | S:
        ...

    def get_last(self, default: S | None = None) -> T | S | None:
        """Returns the last record in the RichSet
        or default value (None) if the RichSet is empty."""
        if self.records:
            return self.records[-1]
        return default

    def last(self) -> T:
        """Returns the last record in the RichSet."""
        if self.records:
            return self.records[-1]
        raise IndexError("RichSet is empty")

    def nth(self, index: int) -> T:
        """Returns the record at the given index."""
        if index < self.size():
            return self.records[index]
        raise IndexError("index out of range")

    @overload
    def get_nth(self, index: int) -> T | None:
        ...

    @overload
    def get_nth(self, index: int, default: S) -> T | S:
        ...

    def get_nth(self, index: int, default: S | None = None) -> T | S | None:
        """Returns the record at the given index
        or default value (None) if the index is out of range."""
        if index < self.size():
            return self.records[index]
        return default

    def one(self) -> T:
        """Returns the one record in the RichSet.

        Currently this method is exactly equivalent to `first()`.
        But this method is intended to be used in uncertain order.
        So this method might not returns not the first record in the future.
        """
        return self.first()

    @overload
    def get_one(self) -> T | None:
        ...

    @overload
    def get_one(self, default: S) -> T | S:
        ...

    def get_one(self, default: S | None = None) -> T | S | None:
        """Returns the one record in the RichSet
        or default value (None) if the RichSet is empty.

        See also `one()`."""
        return self.get_first(default)

    # list manipulations

    def filter(self, f: Callable[[T], bool]) -> RichSet[T]:
        """Returns a new RichSet with filtered records."""
        return RichSet.from_list(list(filter(f, self.records)))

    def unique(self, key: Callable[[T], Key]) -> RichSet[T]:
        """Returns a new RichSet with unique records."""
        new_records = []
        seen = set()
        for r in self.records:
            key_ = key(r)
            if key_ not in seen:
                new_records.append(r)
                seen.add(key_)
        return RichSet.from_list(new_records)

    def map(self, f: Callable[[T], S]) -> RichSet[S]:
        """Returns a new RichSet with mapped records."""
        return RichSet.from_list(list(map(f, self.records)))

    def reduce(
        self,
        fn: Callable[[S, T], S],
        initial: S,
    ) -> S:
        """Returns a reduced value."""
        return functools.reduce(fn, self.records[:], initial)

    def slice(self, start: int, stop: int) -> RichSet[T]:
        """Returns a new RichSet with sliced records."""
        return RichSet.from_tuple(self.records[start:stop])

    def divide_at(self, index: int) -> tuple[RichSet[T], RichSet[T]]:
        """Returns a tuple of two RichSets,
        where the first contains records before the index,
        and the second contains records after the index."""
        return (
            self.slice(0, index),
            self.slice(index, self.size()),
        )

    def pushed(self, record: T) -> RichSet[T]:
        """Returns a new RichSet with the given record pushed to the end."""
        return RichSet.from_tuple(self.records + (record,))

    def pushed_all(self, records: Iterable[T]) -> RichSet[T]:
        """Returns a new RichSet with the given records pushed to the end."""
        return RichSet.from_tuple(self.records + tuple(records))

    def unshifted(self, record: T) -> RichSet[T]:
        """Returns a new RichSet with the given record \
unshifted to the beginning."""
        return RichSet.from_tuple((record,) + self.records)

    def unshifted_all(self, records: Iterable[T]) -> RichSet[T]:
        """Returns a new RichSet with the given records \
unshifted to the beginning."""
        return RichSet.from_tuple(tuple(records) + self.records)

    def popped(self) -> tuple[T, RichSet[T]]:
        """Returns a tuple of the popped record and a new RichSet."""
        if self.is_empty():
            raise IndexError("pop from empty RichSet")
        copied = self.to_list()
        return copied.pop(), RichSet.from_list(copied)

    def popped_n(self, n: int) -> tuple[RichSet[T], RichSet[T]]:
        """Returns a tuple of the popped records and a new RichSet.

        similar to divide_at, but popped records are reversed."""
        if self.size() < n:
            raise IndexError("pop more than size")
        remains, popped_r = self.divide_at(-n)
        return popped_r.reversed(), remains

    def shifted(self) -> tuple[T, RichSet[T]]:
        """Returns a tuple of the shifted record and a new RichSet."""
        if self.is_empty():
            raise IndexError("shift from empty RichSet")
        copied = self.to_list()
        return copied.pop(0), RichSet.from_list(copied)

    def shifted_n(self, n: int) -> tuple[RichSet[T], RichSet[T]]:
        """Returns a tuple of the shifted records and a new RichSet.

        (same as divide_at(n))"""
        if self.size() < n:
            raise IndexError("shift more than size")
        return self.divide_at(n)

    # search

    def index_of(
        self,
        predicate: Callable[[T], bool],
        *,
        reverse: bool = False,
    ) -> int:
        """Returns the index of the first record satisfying the predicate.

        returns -1 if no record satisfies the predicate."""
        if reverse:
            size = self.size()
            for i, item in enumerate(reversed(self.records)):
                if predicate(item):
                    return size - i - 1
            return -1

        for i, r in enumerate(self.records):
            if predicate(r):
                return i
        return -1

    def indices_of(self, predicate: Callable[[T], bool]) -> list[int]:
        """Returns a list of indices of records satisfying the predicate."""
        return [i for i, r in enumerate(self.records) if predicate(r)]

    def search_first(
        self,
        predicate: Callable[[T], bool],
    ) -> tuple[int, T | None]:
        """Returns the first record satisfying the predicate.

        if no record satisfies the predicate, returns (-1, None)."""
        idx = self.index_of(predicate)
        if idx == -1:
            return (-1, None)
        return (idx, self.records[idx])

    def search_last(
        self,
        predicate: Callable[[T], bool],
    ) -> tuple[int, T | None]:
        """Returns the last record satisfying the predicate.

        if no record satisfies the predicate, returns (-1, None)."""
        idx = self.index_of(predicate, reverse=True)
        if idx == -1:
            return (-1, None)
        return (idx, self.records[idx])

    def search_all(
        self,
        predicate: Callable[[T], bool],
    ) -> list[tuple[int, T]]:
        """Returns a list of tuples of indices and records
        satisfying the predicate.

        returns an empty list if no record satisfies the predicate."""
        indices = self.indices_of(predicate)
        return [(i, self.records[i]) for i in indices]

    # set operations

    def union(self, other: RichSet[T]) -> RichSet[T]:
        """Returns a new RichSet with the union of the records."""
        return RichSet.from_list(list(set(self.records) | set(other.records)))

    def intersection(self, other: RichSet[T]) -> RichSet[T]:
        """Returns a new RichSet with the intersection of the records."""
        return RichSet.from_list(list(set(self.records) & set(other.records)))

    def difference(self, other: RichSet[T]) -> RichSet[T]:
        """Returns a new RichSet with the difference of the records."""
        return RichSet.from_list(list(set(self.records) - set(other.records)))

    def symmetric_difference(self, other: RichSet[T]) -> RichSet[T]:
        """Returns a new RichSet with the \
symmetric difference of the records."""
        return RichSet.from_list(list(set(self.records) ^ set(other.records)))

    def is_subset(self, other: RichSet[T]) -> bool:
        """Returns True if self is a subset of other."""
        return set(self.records).issubset(set(other.records))

    def is_superset(self, other: RichSet[T]) -> bool:
        """Returns True if self is a superset of other."""
        return set(self.records).issuperset(set(other.records))

    def is_disjoint(self, other: RichSet[T]) -> bool:
        """Returns True if self and other are disjoint."""
        return set(self.records).isdisjoint(set(other.records))

    def is_equal_as_set(self, other: RichSet[T]) -> bool:
        """Returns True if self and other are same set."""
        return set(self.records) == set(other.records)

    # sortings

    def sorted(
        self,
        *,
        key: Callable[[T], Comparable],
        reverse: bool = False,
    ) -> RichSet[T]:
        """Returns a new RichSet sorted bythe given key."""
        sorted_ = tuple(sorted(self.records, key=key, reverse=reverse))
        return RichSet.from_tuple(sorted_)

    def reversed(self) -> RichSet[T]:
        """Returns a new RichSet with reversed records."""
        return RichSet.from_tuple(self.records[::-1])

    # statistics

    def is_empty(self) -> bool:
        """Returns True if the RichSet is empty."""
        return not self.records

    def is_non_empty(self) -> bool:
        """Returns True if the RichSet is non-empty."""
        return not self.is_empty()

    def size(self) -> int:
        """Returns the number of records in the RichSet."""
        return len(self.records)

    def count(self, predicate: Callable[[T], bool]) -> int:
        """Returns the number of records satisfying the predicate."""
        return sum(1 for r in self.records if predicate(r))



from dataclasses import dataclass

from richset import RichSet


@dataclass(frozen=True)
class Something:
    id: int
    name: str


def test_richset_sorted() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(3, "three"),
            Something(2, "two"),
        ]
    )
    assert rs.sorted(key=lambda r: r.id).to_list() == [
        Something(1, "one"),
        Something(2, "two"),
        Something(3, "three"),
    ]

    assert rs.sorted(key=lambda r: r.name).to_list() == [
        Something(1, "one"),
        Something(3, "three"),
        Something(2, "two"),
    ]

    assert rs.sorted(key=lambda r: r.id, reverse=True).to_list() == [
        Something(3, "three"),
        Something(2, "two"),
        Something(1, "one"),
    ]


def test_richset_reversed() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(3, "three"),
            Something(2, "two"),
        ]
    )
    assert rs.reversed().to_list() == [
        Something(2, "two"),
        Something(3, "three"),
        Something(1, "one"),
    ]



from dataclasses import dataclass

from richset import RichSet


@dataclass(frozen=True)
class Something:
    id: int
    name: str


def test_richset_filter() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "two"),
        ]
    )
    assert rs.filter(lambda r: r.id > 1).to_list() == [
        Something(2, "two"),
        Something(3, "two"),
    ]


def test_richset_unique() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(1, "one"),
        ]
    )
    assert rs.unique(lambda r: r.id).to_list() == [
        Something(1, "one"),
        Something(2, "two"),
    ]


def test_richset_map() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.map(lambda r: r.id).to_list() == [1, 2]


def test_richset_reduce() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
            Something(4, "four"),
        ]
    )
    assert rs.reduce(lambda acc, r: acc + r.id, 0) == 10
    assert rs.reduce(lambda acc, r: acc * r.id, 1) == 24
    assert rs.reduce(lambda acc, r: acc + r.name, "") == "onetwothreefour"



from dataclasses import dataclass

import pytest

from richset import RichSet


@dataclass(frozen=True)
class Something:
    id: int
    name: str


def test_richset_get_first() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.get_first() == Something(1, "one")
    assert rs.get_first(Something(-1, "default")) == Something(1, "one")
    rs2 = RichSet[Something].from_list([])
    assert rs2.get_first() is None
    assert rs2.get_first(Something(-1, "default")) == Something(-1, "default")


def test_richset_first() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.first() == Something(1, "one")

    with pytest.raises(IndexError):
        RichSet.from_list([]).first()


def test_richset_get_last() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.get_last() == Something(2, "two")
    assert rs.get_last(Something(-1, "default")) == Something(2, "two")
    rs2 = RichSet[Something].from_list([])
    assert rs2.get_last() is None
    assert rs2.get_last(Something(-1, "default")) == Something(-1, "default")


def test_richset_last() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.last() == Something(2, "two")

    with pytest.raises(IndexError):
        RichSet.from_list([]).last()


def test_richset_nth() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.nth(0) == Something(1, "one")
    assert rs.nth(1) == Something(2, "two")
    with pytest.raises(IndexError):
        rs.nth(2)

    with pytest.raises(IndexError):
        RichSet.from_list([]).nth(0)


def test_richset_get_nth() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.get_nth(0) == Something(1, "one")
    assert rs.get_nth(1) == Something(2, "two")
    assert rs.get_nth(2, Something(-1, "default")) == Something(-1, "default")
    rs2 = RichSet[Something].from_list([])
    assert rs2.get_nth(0) is None
    assert rs2.get_nth(1, Something(-1, "default")) == Something(-1, "default")


def test_richset_one() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.one() == Something(1, "one")
    with pytest.raises(IndexError) as err:
        RichSet.from_list([]).one()
    assert str(err.value) == "RichSet is empty"


def test_richset_get_one() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.get_one() == Something(1, "one")
    assert rs.get_one(Something(-1, "default")) == Something(1, "one")
    rs2 = RichSet[Something].from_list([])
    assert rs2.get_one() is None
    assert rs2.get_one(Something(-1, "default")) == Something(-1, "default")



from dataclasses import dataclass

from richset import RichSet


@dataclass(frozen=True)
class Something:
    id: int
    name: str


def test_richset_iter() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    for i, r in enumerate(rs):
        if i == 0:
            assert r == Something(1, "one")
        elif i == 1:
            assert r == Something(2, "two")


def test_richset_len() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert len(rs) == 2
    assert len(RichSet[str].from_empty()) == 0



from dataclasses import dataclass

import pytest

from richset import RichSet


@dataclass(frozen=True)
class Something:
    id: int
    name: str


@dataclass  # not immutable (not hashable)
class SomethingElse:
    id: int
    name: str


def test_richset_from_empty() -> None:
    rs = RichSet[str].from_empty()
    assert rs.is_empty()
    assert not rs.is_non_empty()


def test_richset_from_list() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs == RichSet.from_list([Something(1, "one"), Something(2, "two")])


def test_richset_to_list() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.to_list() == [Something(1, "one"), Something(2, "two")]


def test_richset_from_tuple() -> None:
    rs = RichSet.from_tuple(
        (
            Something(1, "one"),
            Something(2, "two"),
        )
    )
    assert rs.to_list() == [Something(1, "one"), Something(2, "two")]


def test_richset_to_tuple() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.to_tuple() == (Something(1, "one"), Something(2, "two"))


def test_richset_from_set() -> None:
    rs = RichSet.from_set(
        {
            Something(1, "one"),
            Something(2, "two"),
        }
    )
    assert rs.to_set() == {Something(1, "one"), Something(2, "two")}


def test_richset_from_frozenset() -> None:
    rs = RichSet.from_frozenset(
        frozenset(
            {
                Something(1, "one"),
                Something(2, "two"),
            }
        )
    )
    assert rs.to_set() == {Something(1, "one"), Something(2, "two")}


def test_richset_to_set() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.to_set() == {Something(1, "one"), Something(2, "two")}
    assert RichSet[Something].from_empty().to_set() == set({})

    rs2 = RichSet.from_list(
        [
            SomethingElse(1, "one"),
            SomethingElse(2, "two"),
        ]
    )
    with pytest.raises(TypeError) as err:
        rs2.to_set()
    assert "non-hashable record:" in str(err.value)


def test_richset_to_frozenset() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.to_frozenset() == frozenset(
        {Something(1, "one"), Something(2, "two")}
    )
    assert RichSet[Something].from_empty().to_frozenset() == set({})

    rs2 = RichSet.from_list(
        [
            SomethingElse(1, "one"),
            SomethingElse(2, "two"),
        ]
    )
    with pytest.raises(TypeError) as err:
        rs2.to_frozenset()
    assert "non-hashable record:" in str(err.value)


def test_richset_to_dict() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.to_dict(lambda r: r.id) == {
        1: Something(1, "one"),
        2: Something(2, "two"),
    }

    rs = RichSet.from_list(
        [
            Something(1, "john"),
            Something(2, "jane"),
            Something(3, "john"),  # name duplicated
        ]
    )

    with pytest.raises(ValueError) as err:
        rs.to_dict(lambda r: r.name)
    assert str(err.value) == "duplicate keys"

    with pytest.raises(ValueError) as err:
        rs.to_dict(lambda r: r.name, duplicated="error")
    assert str(err.value) == "duplicate keys"

    assert rs.to_dict(lambda r: r.name, duplicated="first") == {
        "john": Something(1, "john"),
        "jane": Something(2, "jane"),
    }

    assert rs.to_dict(lambda r: r.name, duplicated="last") == {
        "john": Something(3, "john"),
        "jane": Something(2, "jane"),
    }

    with pytest.raises(ValueError) as err:
        rs.to_dict(lambda r: r.name, duplicated="invalid")  # type: ignore
    assert str(err.value) == "invalid duplicated value"


def test_richset_to_dict_of_list() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "john"),
            Something(2, "jane"),
            Something(3, "john"),  # name duplicated
        ]
    )
    assert rs.to_dict_of_list(lambda r: r.name) == {
        "john": [Something(1, "john"), Something(3, "john")],
        "jane": [Something(2, "jane")],
    }



from dataclasses import dataclass

from richset import RichSet


@dataclass(frozen=True)
class Something:
    id: int
    name: str


def test_richset_union() -> None:
    rs1 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    rs2 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(3, "three"),
        ]
    )
    assert rs1.union(rs2).to_set() == {
        Something(1, "one"),
        Something(2, "two"),
        Something(3, "three"),
    }
    assert rs1.union(rs2).to_set() == rs2.union(rs1).to_set()


def test_richset_intersection() -> None:
    rs1 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    rs2 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(3, "three"),
        ]
    )
    assert rs1.intersection(rs2).to_set() == {Something(1, "one")}
    assert rs1.intersection(rs2).to_set() == rs2.intersection(rs1).to_set()


def test_richset_difference() -> None:
    rs1 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    rs2 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(3, "three"),
        ]
    )
    assert rs1.difference(rs2).to_set() == {Something(2, "two")}
    assert rs2.difference(rs1).to_set() == {Something(3, "three")}


def test_richset_symmetric_difference() -> None:
    rs1 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    rs2 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(3, "three"),
        ]
    )
    assert rs1.symmetric_difference(rs2).to_set() == {
        Something(2, "two"),
        Something(3, "three"),
    }
    assert (
        rs1.symmetric_difference(rs2).to_set()
        == rs2.symmetric_difference(rs1).to_set()
    )


def test_richset_is_subset() -> None:
    rs1 = RichSet.from_list(
        [
            Something(1, "one"),
        ]
    )
    rs2 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(3, "three"),
        ]
    )
    assert rs1.is_subset(rs2)
    assert not rs2.is_subset(rs1)
    assert RichSet[Something].from_empty().is_subset(rs1)
    assert RichSet[Something].from_empty().is_subset(rs2)


def test_richset_is_superset() -> None:
    rs1 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(3, "three"),
        ]
    )
    rs2 = RichSet.from_list(
        [
            Something(1, "one"),
        ]
    )
    assert rs1.is_superset(rs2)
    assert not rs2.is_superset(rs1)
    assert rs1.is_superset(RichSet[Something].from_empty())
    assert rs2.is_superset(RichSet[Something].from_empty())


def test_richset_is_disjoint() -> None:
    rs1 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(3, "three"),
        ]
    )
    rs2 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    rs3 = RichSet.from_list(
        [
            Something(3, "three"),
            Something(4, "four"),
        ]
    )
    assert not rs1.is_disjoint(rs2)
    assert not rs2.is_disjoint(rs1)
    assert not rs1.is_disjoint(rs3)
    assert not rs3.is_disjoint(rs1)
    assert rs2.is_disjoint(rs3)
    assert rs3.is_disjoint(rs2)

    assert RichSet[Something].from_empty().is_disjoint(rs1)
    assert RichSet[Something].from_empty().is_disjoint(rs2)
    assert RichSet[Something].from_empty().is_disjoint(rs3)

    assert rs1.is_disjoint(RichSet[Something].from_empty())
    assert rs2.is_disjoint(RichSet[Something].from_empty())
    assert rs3.is_disjoint(RichSet[Something].from_empty())


def test_richset_is_equal_as_set() -> None:
    rs1 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    rs2 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    rs3 = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )
    assert rs1.is_equal_as_set(rs2)
    assert rs2.is_equal_as_set(rs1)
    assert not rs1.is_equal_as_set(rs3)
    assert not rs3.is_equal_as_set(rs1)

    assert not RichSet[Something].from_empty().is_equal_as_set(rs1)
    assert not RichSet[Something].from_empty().is_equal_as_set(rs2)
    assert not RichSet[Something].from_empty().is_equal_as_set(rs3)

    assert not rs1.is_equal_as_set(RichSet[Something].from_empty())
    assert not rs2.is_equal_as_set(RichSet[Something].from_empty())
    assert not rs3.is_equal_as_set(RichSet[Something].from_empty())



from typing import Protocol, TypeVar, Union

T_contra = TypeVar("T_contra", contravariant=True)


class SupportsDunderLT(Protocol[T_contra]):  # pragma: no cover
    def __lt__(self, __other: T_contra) -> bool:
        ...


class SupportsDunderGT(Protocol[T_contra]):  # pragma: no cover
    def __gt__(self, __other: T_contra) -> bool:
        ...


Comparable = Union[SupportsDunderLT, SupportsDunderGT]



from dataclasses import dataclass

from richset import RichSet


@dataclass(frozen=True)
class Something:
    id: int
    name: str


def test_richset_index_of() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
            Something(4, "four"),
            Something(5, "five"),
        ]
    )
    assert rs.index_of(lambda r: r.id == 1) == 0
    assert rs.index_of(lambda r: r.id == 2) == 1
    assert rs.index_of(lambda r: r.id == 6) == -1
    assert rs.index_of(lambda r: r.id % 2 == 0) == 1
    assert rs.index_of(lambda r: r.id % 2 == 1) == 0
    assert rs.index_of(lambda r: r.id == 10) == -1

    assert rs.index_of(lambda r: r.id == 1, reverse=False) == 0
    assert rs.index_of(lambda r: r.id == 2, reverse=False) == 1
    assert rs.index_of(lambda r: r.id == 6, reverse=False) == -1
    assert rs.index_of(lambda r: r.id % 2 == 0, reverse=False) == 1
    assert rs.index_of(lambda r: r.id % 2 == 1, reverse=False) == 0
    assert rs.index_of(lambda r: r.id == 10, reverse=False) == -1

    assert rs.index_of(lambda r: r.id == 1, reverse=True) == 0
    assert rs.index_of(lambda r: r.id == 2, reverse=True) == 1
    assert rs.index_of(lambda r: r.id == 6, reverse=True) == -1
    assert rs.index_of(lambda r: r.id % 2 == 0, reverse=True) == 3
    assert rs.index_of(lambda r: r.id % 2 == 1, reverse=True) == 4
    assert rs.index_of(lambda r: r.id == 10, reverse=True) == -1


def test_richset_indices() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
        ]
    )
    assert rs.indices_of(lambda r: r.id == 1) == [0]
    assert rs.indices_of(lambda r: r.id == 2) == [1]
    assert rs.indices_of(lambda r: r.id == 3) == []


def test_richset_search_first() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )
    assert rs.search_first(lambda r: r.id == 0) == (-1, None)
    assert rs.search_first(lambda r: r.id == 1) == (0, Something(1, "one"))
    assert rs.search_first(lambda r: r.id == 2) == (1, Something(2, "two"))
    assert rs.search_first(lambda r: r.id == 3) == (2, Something(3, "three"))
    assert rs.search_first(lambda r: r.id > 1) == (1, Something(2, "two"))


def test_richset_search_last() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )
    assert rs.search_last(lambda r: r.id == 0) == (-1, None)
    assert rs.search_last(lambda r: r.id == 1) == (0, Something(1, "one"))
    assert rs.search_last(lambda r: r.id == 2) == (1, Something(2, "two"))
    assert rs.search_last(lambda r: r.id == 3) == (2, Something(3, "three"))
    assert rs.search_last(lambda r: r.id > 1) == (2, Something(3, "three"))


def test_richset_search_all() -> None:
    rs = RichSet.from_list(
        [
            Something(1, "one"),
            Something(2, "two"),
            Something(3, "three"),
        ]
    )
    assert rs.search_all(lambda r: r.id == 0) == []
    assert rs.search_all(lambda r: r.id == 1) == [(0, Something(1, "one"))]
    assert rs.search_all(lambda r: r.id == 2) == [(1, Something(2, "two"))]
    assert rs.search_all(lambda r: r.id > 1) == [
        (1, Something(2, "two")),
        (2, Something(3, "three")),
    ]
